# Purpose: Split a string into sentences, based on a set of terminators.
#          This is a helper class for ooba_client.py.
import pysbd
import typing


class SentenceSplitter:
    '''
    Purpose: Split an English string into sentences.
    '''

    # anything that can't be in a real response
    END_OF_INPUT = ''

    def __init__(self):
        self.printed_idx = 0
        self.full_response = ''
        self.segmenter = pysbd.Segmenter(
            language="en", clean=False, char_span=True)

    def by_sentence(self, new_token: str) \
            -> typing.Generator[str, None, None]:
        '''
        Collects tokens into a single string, looks for ends of english
        sentences, then yields each sentence as soon as it's found.

        Parameters:
            new_token: str, the next token to add to the string

        Returns:
            Generator[str, None, None], yields each sentence

        Note:
        When there is no longer any input, the caller must pass
        SentenceSplitter.END_OF_INPUT to this function.  This
        function will then yield any remaining text, even if it
        doesn't look like a full sentence.
        '''

        self.full_response += new_token
        unseen = self.full_response[self.printed_idx:]

        # if we've reached the end of input, yield it all,
        # even if we don't think it's a full sentence.
        if (self.END_OF_INPUT == new_token):
            to_print = unseen.strip()
            if (to_print):
                yield unseen
            self.printed_idx += len(unseen)
            return

        segments = self.segmenter.segment(unseen)

        # any remaining non-sentence things will be in the last element
        # of the list.  Don't print that yet.  At the very worst, we'll
        # print it when the END_OF_INPUT signal is reached.
        for sentence_w_char_spans in segments[:-1]:

            # sentence_w_char_spans is a class with the following fields:
            #  - sent: str, sentence text
            #  - start: start idx of 'sent', relative to original string
            #  - end: end idx of 'sent', relative to original string
            #
            # we want to remove the last '\n' if there is one.
            # we do want to include any other whitespace, though.

            to_print = sentence_w_char_spans.sent  # type: ignore
            if (to_print.endswith('\n')):
                to_print = to_print[:-1]

            yield to_print

        # since we've printed all the previous segments,
        # the start of the last segment becomes the starting
        # point for the next roud.
        if (len(segments) > 0):
            self.printed_idx += segments[-1].start  # type: ignore



# Purpose: Discord client for Rosie
#

import discord
import re
import textwrap
import typing

from oobabot.fancy_logging import get_logger
from oobabot.ooba_client import OobaClient
from oobabot.response_stats import AggregateResponseStats


# only accept english keyboard characters in messages
# if other ones appear, they will be filtered out
FORBIDDEN_CHARACTERS = r'[^a-zA-Z0-9\-\\=\[\];,./~!@#$%^&*()_+{}|:"<>?`\' ]'
FORBIDDEN_CHARACTERS_PATTERN = re.compile(FORBIDDEN_CHARACTERS)


def sanitize_string(raw_string: str) -> str:
    '''
    Filter out any characters that are not commonly on a
    US-English keyboard
    '''
    return FORBIDDEN_CHARACTERS_PATTERN.sub('', raw_string)


def sanitize_message(raw_message: discord.Message) -> dict[str, str]:
    author = sanitize_string(raw_message.author.name)

    raw_guild = raw_message.guild
    if raw_guild:
        raw_guild_name = raw_guild.name
    else:
        raw_guild_name = 'DM'

    return {
        'author': author,
        'author_shortname': author.split('#')[0],
        'message_text': sanitize_string(raw_message.content).strip(),
        'server': sanitize_string(raw_guild_name),
    }


class PromptGenerator:
    '''
    Purpose: generate a prompt for the AI to use, given
    the message history and persona.
    '''

    # this is set by the AI, and is the maximum length
    # it will understand before it starts to ignore
    # the rest of the prompt
    MAX_PROMPT_LEN = 2048

    def __init__(self, ai_name: str,
                 ai_persona: str):
        self.ai_name = ai_name
        self.ai_persona = ai_persona

    async def generate_prompt(
            self,
            ai_user_id: int,
            message_history: typing.AsyncIterator[discord.Message]) \
            -> str:
        '''
        Generates a prompt for the AI to use based on the message.
        '''

        prompt = textwrap.dedent(f'''
        You are in a chat room with multiple participants.
        Below is a transcript of recent messages in the conversation.
        Write the next message that you would send in this conversation,
        from the point of view of the participant named "{self.ai_name}".
        Here is some background information about "{self.ai_name}":
        {self.ai_persona}
        All responses you write must be from the point of view of
        {self.ai_name}, and plausible for {self.ai_name} to say in
        this conversation.  Do not continue the conversation as
        anyone other than {self.ai_name}.
        ### Transcript:
        ''')

        # put this at the very end to tell the UI what it
        # should complete.  But generate this now so we
        # know how long it is.
        prompt_footer = f'{self.ai_name} says:\n'

        # add on more history, but only if we have room
        # if we don't have room, we'll just truncate the history
        # by discarding the oldest messages first
        # this is s
        # it will understand before ignore
        #
        prompt_len_remaining = self.MAX_PROMPT_LEN - \
            len(prompt) - len(prompt_footer)

        # history_lines is newest first, so figure out
        # how many we can take, then append them in
        # reverse order
        added_history = 0
        history_lines = []
        async for raw_message in message_history:
            clean_message = sanitize_message(raw_message)

            author = clean_message["author"]
            if raw_message.author.id == ai_user_id:
                author = self.ai_name

            if clean_message["message_text"] is None:
                continue

            line = f'{author} says:\n' + \
                f'{clean_message["message_text"]}\n\n'

            if len(line) > prompt_len_remaining:
                get_logger().warn(
                    'prompt too long, truncating history.  ' +
                    f'added {added_history} lines, dropped ' +
                    f'{len(history_lines)} lines'
                )
                break

            prompt_len_remaining -= len(line)
            history_lines.append(line)
            added_history += 1

        history_lines.reverse()
        prompt += ''.join(history_lines)

        prompt += prompt_footer

        return prompt


class DiscordBot(discord.Client):
    HIST_SIZE = 10

    def __init__(self,
                 ooba_client: OobaClient,
                 ai_name: str,
                 ai_persona: str,
                 wakewords: list[str],
                 log_all_the_things: bool):
        self.ooba_client = ooba_client

        self.ai_name = ai_name
        self.ai_persona = ai_persona
        self.ai_user_id = -1
        self.wakewords = wakewords
        self.log_all_the_things = log_all_the_things

        self.average_stats = AggregateResponseStats(ooba_client)
        self.prompt_generator = PromptGenerator(ai_name, ai_persona)

        # match messages that include any `wakeword`, but not as part of
        # another word
        self.wakeword_patterns = [
            re.compile(f'\\b{wakeword}\\b', re.IGNORECASE)
            for wakeword in wakewords
        ]

        intents = discord.Intents.default()
        intents.message_content = True

        super().__init__(intents=intents)

    async def on_ready(self) -> None:
        guilds = self.guilds
        num_guilds = len(guilds)
        num_channels = sum([len(guild.channels) for guild in guilds])

        if self.user:
            self.ai_user_id = self.user.id
            user_id_str = str(self.ai_user_id)
        else:
            user_id_str = '<unknown>'

        get_logger().info(
            f'Connected to discord as {self.user} (ID: {user_id_str})')
        get_logger().debug(
            f'monitoring DMs, plus {num_channels} channels across ' +
            f'{num_guilds} server(s)')

        get_logger().debug(f'AI name: {self.ai_name}')
        get_logger().debug(f'AI persona: {self.ai_persona}')

        str_wakewords = ", ".join(
            self.wakewords) if self.wakewords else "<none>"
        get_logger().debug(
            f'wakewords: {str_wakewords}'
        )

    def run(self, token: str) -> None:
        super().run(token)

    def should_reply_to_message(self, message: discord.Message) -> bool:
        # we do not want the bot to reply to itself
        if self.user and message.author.id == self.user.id:
            return False

        # reply to all private messages
        if discord.ChannelType.private == message.channel.type:
            return True

        # reply to all messages that include a wakeword
        for wakeword_pattern in self.wakeword_patterns:
            if wakeword_pattern.search(message.content):
                return True

        # reply to all messages in which we're @-mentioned
        if self.user and self.user.id in [m.id for m in message.mentions]:
            return True

        # ignore anything else
        return False

    def log_stats(self) -> None:
        self.average_stats.write_stat_summary_to_log()

    async def on_message(self, raw_message: discord.Message) -> None:
        if not self.should_reply_to_message(raw_message):
            return
        try:
            async with raw_message.channel.typing():
                await self.send_response(raw_message)
        except Exception as e:
            get_logger().error(
                f'Exception while sending response: {e}', exc_info=True)

    async def send_response(self, raw_message: discord.Message) -> None:
        clean_message = sanitize_message(raw_message)
        author = clean_message['author']
        server = clean_message['server']
        get_logger().debug(
            f'Request from {author} in server [{server}]')

        recent_messages = raw_message.channel.history(limit=self.HIST_SIZE)

        prompt = await self.prompt_generator.generate_prompt(
            self.ai_user_id, recent_messages)

        response_stats = self.average_stats.log_request_arrived(prompt)
        if self.log_all_the_things:
            print('Prompt:\n----------\n')
            print(prompt)
            print('Response:\n----------\n')

        try:
            async for sentence in self.ooba_client.request_by_sentence(
                prompt
            ):
                if self.log_all_the_things:
                    print(sentence)

                # if the AI gives itself a second line, just ignore
                # the line instruction and continue
                if f'{self.ai_name} says:' == sentence:
                    get_logger().warning(
                        f'Filtered out "{sentence}" from response, continuing'
                    )
                    continue

                # hack: abort response if it looks like the AI is
                # continuing the conversation as someone else
                if sentence.endswith(' says:'):
                    get_logger().warning(
                        f'Filtered out "{sentence}" from response, aborting'
                    )
                    break

                await raw_message.channel.send(sentence)
                response_stats.log_response_part()

        except Exception as err:
            get_logger().error(f'Error: {str(err)}')
            self.average_stats.log_response_failure()
            return

        response_stats.write_to_log(f"Response to {author} done!  ")
        self.average_stats.log_response_success(response_stats)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#

import asyncio
import signal
import sys

from oobabot.discord_bot import DiscordBot
from oobabot.fancy_logging import init_logging
from oobabot.ooba_client import OobaClient
from oobabot.settings import Settings


class LocalREPL:
    # local REPL for testing
    def __init__(self, ooba_client):
        self.ooba_client = ooba_client

    def show_prompt(self) -> None:
        print("\n>>> ", end='', flush=True)

    async def start(self) -> None:
        self.show_prompt()
        for user_prompt in sys.stdin:
            async for token in self.ooba_client.request_by_token(
                user_prompt
            ):
                if token:
                    print(token, end='', flush=True)
                else:
                    # end of response
                    print('')
            self.show_prompt()


def main():

    logger = init_logging()
    settings = Settings()

    bot = None

    def sigint_handler(_signum, _frame):
        logger.info('Received SIGINT, exiting...')
        if bot:
            bot.log_stats()
        exit(1)

    signal.signal(signal.SIGINT, sigint_handler)

    ooba_client = OobaClient(settings.base_url)

    logger.debug(f'Oobabooga base URL: {settings.base_url}')
    connect_error_msg = asyncio.run(ooba_client.try_connect())
    if connect_error_msg:
        logger.error(
            f'Could not connect to ooba server: [{ooba_client.api_url}]')
        logger.error('Please check the URL and try again.')
        logger.error(f'Reason: {connect_error_msg}')
        sys.exit(1)

    logger.info('Connected to Oobabooga!')

    if settings.local_repl:
        logger.debug('Using local REPL, not connecting to Discord')
        coroutine = LocalREPL(ooba_client).start()
    else:
        logger.debug('Connecting to Discord... ')
        bot = DiscordBot(
            ooba_client,
            ai_name=settings.ai_name,
            ai_persona=settings.persona,
            wakewords=settings.wakewords,
            log_all_the_things=settings.log_all_the_things)
        coroutine = bot.start(settings.DISCORD_TOKEN)

    asyncio.run(coroutine)



import time

from oobabot.ooba_client import OobaClient
from oobabot.fancy_logging import get_logger


class ResponseStats:
    '''
    Purpose: collects timing and rate statistics for a single response
    '''

    def __init__(self, ooba_client: OobaClient, prompt: str):
        self.ooba_client = ooba_client
        self.start_time = time.time()
        self.start_tokens = ooba_client.total_response_tokens
        self.duration = 0
        self.latency = 0
        self.tokens = 0
        self.prompt_len = len(prompt)

    def log_response_part(self) -> None:
        '''
        Call this each time the response is updated by the AI.
        '''

        now = time.time()
        if not self.latency:
            self.latency = now - self.start_time
        self.duration = now - self.start_time
        self.tokens = self.ooba_client.total_response_tokens - \
            self.start_tokens

    def tokens_per_second(self) -> float:
        '''
        Returns the rate at which tokens were generated, in tokens per second.
        '''
        if not self.duration:
            return 0
        return self.tokens / self.duration

    def write_to_log(self, log_prefix: str) -> None:
        '''
        This writes the statistics for this specific
        request to the log.
        '''
        get_logger().debug(
            log_prefix +
            f"tokens: {self.tokens}, " +
            f"time: {self.duration:.2f}s, " +
            f"latency: {self.latency:.2f}s, " +
            f"rate: {self.tokens_per_second():.2f} tok/s")


class AggregateResponseStats:
    '''
    Purpose: collects timing and rate statistics for all AggregateResponseStats
    '''

    def __init__(self, ooba_client: OobaClient):
        self.ooba_client = ooba_client
        self.total_requests_received = 0
        self.total_successful_responses = 0
        self.total_failed_responses = 0
        self.total_response_time_seconds = 0
        self.total_response_latency_seconds = 0
        self.prompt_max_chars = 0
        self.prompt_min_chars = 0
        self.prompt_total_chars = 0

    def log_request_arrived(self, prompt) -> ResponseStats:
        '''
        Call this when a request has arrived.
        This must be followed by zero or more calls
        to log_response_part(), and then exactly one call to
        either log_response_failure() or log_response_success().
        '''
        result = ResponseStats(self.ooba_client, prompt)

        self.total_requests_received += 1
        # update the prompt stats now
        if self.prompt_max_chars < result.prompt_len:
            self.prompt_max_chars = result.prompt_len
        if self.prompt_min_chars > result.prompt_len:
            self.prompt_min_chars = result.prompt_len
        self.prompt_total_chars += result.prompt_len

        return result

    def log_response_failure(self) -> None:
        '''
        Track the statistics for a failed response.
        '''
        self.total_failed_responses += 1

    def log_response_success(self, response: ResponseStats) -> None:
        '''
        Track the statistics for a successful response, and
        averages them into the overall statistics.

        Parameters:
         - response: the Response object returned
              by log_request_arrived()
        '''
        self.total_successful_responses += 1
        self.total_response_time_seconds += response.duration
        self.total_response_latency_seconds += response.latency

    def error_rate(self) -> float:
        '''
        Returns the percentage of requests that failed.
        '''
        if 0 == self.total_requests_received:
            return 0.0
        return 100 * self.total_failed_responses / self.total_requests_received

    def average_response_time(self) -> float:
        '''
        Returns the average response time in seconds.
        '''
        if 0 == self.total_successful_responses:
            return 0.0
        return self.total_response_time_seconds / \
            self.total_successful_responses

    def average_response_latency(self) -> float:
        '''
        Returns the average response latency in seconds.
        '''
        if 0 == self.total_successful_responses:
            return 0.0
        return self.total_response_latency_seconds / \
            self.total_successful_responses

    def average_tokens_per_second(self) -> float:
        '''
        Returns the average rate at which tokens were generated,
        in tokens per second.
        '''
        if 0 == self.total_successful_responses:
            return 0.0
        return self.ooba_client.total_response_tokens / \
            self.total_response_time_seconds

    def average_prompt_length(self) -> float:
        '''
        Returns the average prompt length in characters.
        '''
        if 0 == self.total_requests_received:
            return 0.0
        return self.prompt_total_chars / self.total_requests_received

    def write_stat_summary_to_log(self) -> None:
        '''
        This writes a summary of the statistics to the log.
        Call this after all AggregateResponseStats have been handled.
        '''
        if 0 == self.total_requests_received:
            get_logger().info('No requests handled')
            return

        get_logger().info(
            f'Recevied {self.total_requests_received} request(s), ' +
            f'sent {self.total_successful_responses} successful responses ' +
            f'and failed to send one {self.total_failed_responses} times(s)')

        if (self.total_failed_responses > 0):
            get_logger().error(
                'Error rate:                  ' +
                f'{self.error_rate()}%'
            )

        if (self.total_successful_responses > 0):
            get_logger().debug(
                'Average response time:       ' +
                f'{self.average_response_time():6.2f}s'
            )
            get_logger().debug(
                'Average response latency:    ' +
                f'{self.average_response_latency():6.2f}s'
            )
            get_logger().debug(
                'Average tokens per response: ' +
                f'{self.average_tokens_per_second():6.2f}'
            )

        if self.total_response_time_seconds > 0:
            get_logger().debug(
                'Average tokens per second:   ' +
                f'{self.average_tokens_per_second():6.2f}'
            )

        get_logger().debug(
            'Prompt length:               ' +
            f'max: {self.prompt_max_chars}, ' +
            f'min: {self.prompt_min_chars}, ' +
            f'avg: {self.average_prompt_length():.2f}'
        )



# Purpose: Streaming client for the Ooba API.
# Can provide the response by token or by sentence.
#

import json
import typing
import websockets as ws  # weird, but needed to avoid long lines later

from oobabot.sentence_splitter import SentenceSplitter
from urllib.parse import urljoin


class OobaClient:
    # Purpose: Streaming client for the Ooba API.
    # Can provide the response by token or by sentence.

    END_OF_INPUT = ''

    def __init__(self, base_url: str):
        self.api_url = urljoin(base_url, self.STREAMING_URI_PATH)
        self.total_response_tokens = 0

    DEFAULT_REQUEST_PARAMS = {
        'max_new_tokens': 250,
        'do_sample': True,
        'temperature': 1.3,
        'top_p': 0.1,
        'typical_p': 1,
        'repetition_penalty': 1.18,
        'top_k': 40,
        'min_length': 0,
        'no_repeat_ngram_size': 0,
        'num_beams': 1,
        'penalty_alpha': 0,
        'length_penalty': 1,
        'early_stopping': False,
        'seed': -1,
        'add_bos_token': True,
        'truncation_length': 2048,
        'ban_eos_token': False,
        'skip_special_tokens': True,
        'stopping_strings': [],
    }

    STREAMING_URI_PATH = './api/v1/stream'

    async def try_connect(self) -> str:
        '''
        Attempt to connect to the oobabooga server.

        Returns:
            str, error message if unsuccessful, empty string on success
        '''
        try:
            async with ws.connect(self.api_url) as _:  # type: ignore
                return ''
        except Exception as e:
            return str(e)

    async def request_by_sentence(self, prompt: str) \
            -> typing.AsyncIterator[str]:
        '''
        Yields each complete sentence of the response as it arrives.
        '''

        splitter = SentenceSplitter()
        async for new_token in self.request_by_token(prompt):
            for sentence in splitter.by_sentence(new_token):
                yield sentence

    async def request_by_token(self, prompt: str) \
            -> typing.AsyncIterator[str]:
        '''
        Yields each token of the response as it arrives.
        '''

        request = {
            'prompt': prompt,
        }
        request.update(self.DEFAULT_REQUEST_PARAMS)

        async with ws.connect(self.api_url) as websocket:  # type: ignore
            await websocket.send(json.dumps(request))

            while True:
                incoming_data = await websocket.recv()
                incoming_data = json.loads(incoming_data)

                if 'text_stream' == incoming_data['event']:
                    if (incoming_data['text']):
                        self.total_response_tokens += 1
                        yield incoming_data['text']

                elif 'stream_end' == incoming_data['event']:
                    # Make sure any unprinted text is flushed.
                    yield self.END_OF_INPUT
                    return




import argparse
import os


class Settings(argparse.ArgumentParser):
    # Purpose: reads settings from environment variables and command line
    #          arguments

    DISCORD_TOKEN_ENV_VAR = 'DISCORD_TOKEN'
    DISCORD_TOKEN = os.environ.get(DISCORD_TOKEN_ENV_VAR, '')

    OOBABOT_PERSONA_ENV_VAR = 'OOBABOT_PERSONA'
    OOBABOT_PERSONA = os.environ.get(OOBABOT_PERSONA_ENV_VAR, '')

    DEFAULT_WAKEWORDS = ['oobabot']
    DEFAULT_URL = 'ws://localhost:5005'

    def __init__(self):
        self._settings = None
        self.wakewords = []

        super().__init__(
            description="Discord bot for oobabooga's text-generation-webui",
            epilog='Also, to authenticate to Discord, you must set the ' +
            'environment variable:\n'
            f"\t{self.DISCORD_TOKEN_ENV_VAR} = <your bot's discord token>"
        )
        self.add_argument(
            '--base-url',
            type=str,
            default=self.DEFAULT_URL,
            help='Base URL for the oobabooga instance.  ' +
            'This should be ws://hostname[:port] for plain websocket ' +
            'connections, or wss://hostname[:port] for websocket ' +
            'connections over TLS.'
        )
        self.add_argument(
            '--ai-name',
            type=str,
            default='oobabot',
            help='Name of the AI to use for requests.  ' +
            'This can be whatever you want, but might make sense ' +
            'to be the name of the bot in Discord.'
        )
        self.add_argument(
            '--wakewords',
            type=str,
            nargs='*',
            default=self.DEFAULT_WAKEWORDS,
            help='One or more words that the bot will listen for.\n ' +
            'The bot will listen in all discord channels can ' +
            'access for one of these words to be mentioned, then reply ' +
            'to any messages it sees with a matching word.  ' +
            'The bot will always reply to @-mentions and ' +
            'direct messages, even if no wakewords are supplied.')
        self.add_argument(
            '--persona',
            type=str,
            default=self.OOBABOT_PERSONA,
            help='This prefix will be added in front of every user-supplied ' +
            "request.  This is useful for setting up a 'character' for the " +
            'bot to play.  Alternatively, this can be set with the ' +
            f'{self.OOBABOT_PERSONA_ENV_VAR} environment variable.'
        )
        self.add_argument(
            '--local-repl',
            default=False,
            help='start a local REPL, instead of connecting to Discord',
            action='store_true'
        )
        self.add_argument(
            '--log-all-the-things',
            default=False,
            help='prints all oobabooga requests and responses in their ' +
            'entirety to STDOUT',
            action='store_true'
        )

    def settings(self) -> dict[str, str]:
        if self._settings is None:
            self._settings = self.parse_args().__dict__

            # this is a bit of a hack, but it's the only setting
            # that's not a string, so with this we can be more
            # strict with the type hints later
            self.wakewords = self._settings.pop('wakewords')

            # either we're using a local REPL, or we're connecting to Discord.
            # assume the user wants to connect to Discord
            if not (self.local_repl or self.DISCORD_TOKEN):
                msg = f"Please set the '{Settings.DISCORD_TOKEN_ENV_VAR}' " + \
                    "environment variable to your bot's discord token."
                # will exit() after printing
                self.error(msg)

        return self._settings

    def __getattr__(self, name) -> str:
        return self.settings().get(name, '')

    def __repr__(self) -> str:
        return super().__repr__()
